#ifndef __MD_LIB_ITERMAP_H__
#define __MD_LIB_ITERMAP_H__

#include <cassert>
#include <string>

#include "MD/MD_Message.h"
#include "MD/MD_DateTime.h"
#include "MD/MD_EpochNS.h"
#include "MD/MD_FieldMap.h"
#include "MD/MD_numeric.h"
#include "MD/MD_Types.h"
#include "MD/MD_Protocols.h"

namespace MD {

class IterMap : public MD_Message {
public:
  typedef MD_MsgIterator<IterMap> iterator;
  class Field;

  IterMap(bool opt = true);
  IterMap(bool opt, const MD_Message::Allocator& allocator);
  IterMap(const IterMap& src);
  IterMap(IterMap&& src);
  IterMap(const unsigned char* ptr, uint32_t size);
  IterMap(const unsigned char* ptr,
          uint32_t size,
          const MD_Message::Allocator& allocator);
  virtual ~IterMap();

  IterMap & operator= (const IterMap & src);

  // Give the user a new empty message of the same type as this one.
  // The user is responsible for deleting the pointer.
  virtual std::unique_ptr<MD::MD_Message> create() const;

  // Give the user a new copy of the message.
  // The user is responsible for deleting the pointer.
  virtual std::unique_ptr<MD::MD_Message> clone() const;

  // Give the user a new pointer containing this message.
  // The user is responsible for deleting the pointer.
  virtual std::unique_ptr<MD_Message> move();

  //! Accessor methods
  virtual bool getChar     (FieldID id, char        &result) const;
  virtual bool getInt      (FieldID id, int         &result) const;
  virtual bool getInt64    (FieldID id, int64_t     &result) const;
  virtual bool getFloat    (FieldID id, float       &result) const;
  virtual bool getDouble   (FieldID id, double      &result) const;
  virtual bool getString   (FieldID id, String      &result) const;
  virtual bool getBuffer   (FieldID id, Buffer      &result) const;
  virtual bool getDateTime (FieldID id, MD_DateTime &result) const;
  virtual bool getEpochNS  (FieldID id, MD_EpochNS  &result) const;

  virtual bool addChar     (FieldID id, char        data);
  virtual bool addInt      (FieldID id, int         data);
  virtual bool addInt64    (FieldID id, int64_t     data);
  virtual bool addFloat    (FieldID id, float       data);
  virtual bool addDouble   (FieldID id, double      data);
  virtual bool addString   (FieldID id, String      data);
  virtual bool addBuffer   (FieldID id, Buffer      data);
  virtual bool addDateTime (FieldID id, MD_DateTime data);
  virtual bool addEpochNS  (FieldID id, MD_EpochNS  data);

  virtual bool updateChar     (FieldID id, char        data);
  virtual bool updateInt      (FieldID id, int         data);
  virtual bool updateInt64    (FieldID id, int64_t     data);
  virtual bool updateFloat    (FieldID id, float       data);
  virtual bool updateDouble   (FieldID id, double      data);
  virtual bool updateString   (FieldID id, String      data);
  virtual bool updateBuffer   (FieldID id, Buffer      data);
  virtual bool updateDateTime (FieldID id, MD_DateTime data);
  virtual bool updateEpochNS  (FieldID id, MD_EpochNS data);

  // Pull in base class convenience methods
  using MD::MD_Message::getString;
  using MD::MD_Message::addString;
  using MD::MD_Message::updateString;

  virtual bool erase (FieldID id);
  virtual bool erase (const std::set<FieldID> &ids);
  virtual bool erase (const std::vector<FieldID> &ids);
  virtual void clear();

  virtual bool empty() const;
  virtual bool exists (FieldID id) const;

  virtual const unsigned char * getBinary() const;
  virtual bool setBinary (const unsigned char * ptr, uint32_t size);

  virtual uint32_t getSize() const;

  virtual MD::PROTOCOLS::MsgProtocol protocol() const;

  virtual const unsigned char *
  getDataPtr (FieldID id, TypeID & type, int & size) const;

  iterator begin() const;
  iterator end() const;

private:
  virtual void getFieldIDs(std::back_insert_iterator<std::vector<FieldID> > inserter) const;

  template <typename FieldType>
  bool addField (FieldID id, FieldType data);

  template <typename FieldType>
  bool updateField (FieldID id, FieldType data);

  template <typename FieldType>
  bool getHelper(FieldID id, FieldType & data) const;

  static const size_t DATA_START = 1;
};

class IterMap::Field
{
  typedef MD::DATATYPES::TypeID TypeID;
  typedef MD::FIELDS::FieldID   FieldID;

  public:
  Field(char * ptr);

  MD_LIB_API FieldID    id() const;
  MD_LIB_API TypeID   type() const;
  MD_LIB_API uint32_t size() const;

  MD_LIB_API char     *data_ptr() const;

  MD_LIB_API operator char()    const;
  MD_LIB_API operator int32_t() const;
  MD_LIB_API operator int64_t() const;
  MD_LIB_API operator float()   const;
  MD_LIB_API operator double()  const;
  MD_LIB_API operator String()  const;
  MD_LIB_API operator Buffer()  const;
  MD_LIB_API operator MD_EpochNS() const;

  bool operator==(const IterMap::Field &rhs) const;
private:
  // Length of the header portion of the field.
  uint8_t header_length()  const;
  // optimization table for size_len().
  static const uint8_t size_lens[];
  char    *ptr()      const;
  char    *next_ptr() const;
  char    *id_ptr()   const;
  char    *size_ptr() const;
  uint8_t  size_len() const;
  bool     isValid()    const;
  void     next();
  char    *m_ptr;
  friend class IterMap;
  friend class MD_MsgIterator<IterMap>;
  friend class MD_Message;
};

// Inline function definitions.

inline IterMap::IterMap(bool opt) : IterMap(opt, MD_Message::DEFAULT_ALLOCATOR)
{}

inline IterMap::IterMap(bool opt, const MD_Message::Allocator& allocator)
    : MD_Message(allocator)
{
    m_data->clear();
    m_pOffsets->clear();

    m_data->resize(1);
    m_data->begin()[0] = 1;
}

inline IterMap::IterMap(const IterMap& src) : MD_Message()
{
    *this = src;
}

inline IterMap::IterMap(IterMap&& src) : MD_Message()
{
    swap(src);
}

inline IterMap::IterMap(const unsigned char* ptr, uint32_t size)
    : IterMap(ptr, size, MD_Message::DEFAULT_ALLOCATOR)
{}

inline IterMap::IterMap(const unsigned char* ptr,
                        uint32_t size,
                        const MD_Message::Allocator& allocator)
{
    setBinary(ptr, size);  // setBinary clears m_data and m_pOffsets
}

inline IterMap::~IterMap() { }

inline std::unique_ptr<MD::MD_Message> IterMap::create() const {
  return std::unique_ptr<MD::MD_Message>(new IterMap);
}

inline std::unique_ptr<MD::MD_Message> IterMap::clone() const {
  return std::unique_ptr<MD::MD_Message>(new IterMap(*this));
}

inline std::unique_ptr<MD::MD_Message> IterMap::move() {
    return std::unique_ptr<MD::MD_Message>(new IterMap(std::move(*this)));
}

inline bool IterMap::getChar(FieldID id, char & result) const
{
  return getHelper(id, result);
}

inline bool IterMap::getInt (FieldID id, int & result) const
{
  return getHelper<int32_t>(id, result);
}

inline bool IterMap::getInt64 (FieldID id, int64_t & result) const {
  return getHelper(id, result);
}

inline bool IterMap::getFloat (FieldID id, float & result) const {
  return getHelper(id, result);
}

inline bool IterMap::getDouble (FieldID id, double & result) const {
  return getHelper(id, result);
}

inline bool IterMap::getString (FieldID id, String & result) const {
  return getHelper(id, result);
}

inline bool IterMap::getBuffer (FieldID id, Buffer & result) const {
  return getHelper(id, result);
}

inline bool IterMap::getDateTime (FieldID id, MD_DateTime & result) const {
  return getHelper(id, result);
}

inline bool IterMap::getEpochNS (FieldID id, MD_EpochNS & result) const {
  return getHelper(id, result);
}

inline bool IterMap::Field::isValid() const { return m_ptr != NULL; }

inline bool IterMap::exists (FieldID id) const {
  return getField<Field>(id).isValid();
}

inline bool IterMap::addChar(FieldID id, char data) {
  return !exists(id) ? addField(id, data) : false;
}

inline bool IterMap::addInt (FieldID id, int data) {
  return !exists(id) ? addField<int32_t>(id, data) : false;
}

inline bool IterMap::addInt64 (FieldID id, int64_t data) {
  return !exists(id) ? addField(id, data) : false;
}

inline bool IterMap::addFloat (FieldID id, float data) {
  return !exists(id) ? addField(id, data) : false;
}

inline bool IterMap::addDouble (FieldID id, double data) {
  return !exists(id) ? addField(id, data) : false;
}

inline bool IterMap::addString (FieldID id, String data) {
  return !exists(id) ? addField(id, data) : false;
}

inline bool IterMap::addBuffer (FieldID id, Buffer data) {
  return !exists(id) ? addField(id, data) : false;
}

inline bool IterMap::addDateTime (FieldID id, MD_DateTime data) {
  return !exists(id) ? addField(id, data) : false;
}

inline bool IterMap::addEpochNS (FieldID id, MD_EpochNS data) {
  return !exists(id) ? addField(id, data) : false;
}

inline bool IterMap::updateChar(FieldID id, char data) {
  return updateField(id, data);
}

inline bool IterMap::updateInt (FieldID id, int data) {
  return updateField<int32_t>(id, data);
}

inline bool IterMap::updateInt64 (FieldID id, int64_t data) {
  return updateField(id, data);
}

inline bool IterMap::updateFloat (FieldID id, float data) {
  return updateField(id, data);
}

inline bool IterMap::updateDouble (FieldID id , double data) {
  return updateField(id, data);
}

inline bool IterMap::updateString (FieldID id, String data) {
  return updateField(id, data);
}

inline bool IterMap::updateBuffer (FieldID id, Buffer data) {
  return updateField(id, data);
}

inline bool IterMap::updateDateTime (FieldID id, MD_DateTime data) {
  return updateField(id, data);
}

inline bool IterMap::updateEpochNS (FieldID id, MD_EpochNS data) {
  return updateField(id, data);
}

inline void IterMap::clear()
{
  m_data->resize(DATA_START);
  m_pOffsets->clear();
}

inline IterMap::iterator IterMap::begin() const
{ return IterMap::iterator(m_data->begin() + DATA_START); }

inline IterMap::iterator IterMap::end() const
{ return IterMap::iterator(m_data->end()); }

inline uint32_t IterMap::getSize() const {
  return (uint32_t)m_data->size();
}

inline bool IterMap::empty() const
{ return getSize() <= DATA_START; }

inline const unsigned char * IterMap::getBinary() const {
  return reinterpret_cast <unsigned char *> (m_data->begin());
}

inline MD::PROTOCOLS::MsgProtocol IterMap::protocol() const {
  return MD::PROTOCOLS::ITERMAP;
}

inline char *IterMap::Field::id_ptr()   const { return m_ptr; }

inline MD::FIELDS::FieldID IterMap::Field::id() const {
  assert(isValid());
  return MD::getint16le(id_ptr());
}

inline MD::DATATYPES::TypeID IterMap::Field::type() const {
  const MD::MD_FieldMap &inst = MD::MD_FieldMap::instance();
  assert(isValid());
  MD::MD_FieldMap::const_iterator i = inst.findById(id());
  return (i != inst.end()) ? i->type : MD::DATATYPES::RAW;
}

inline uint8_t IterMap::Field::size_len() const
{ return size_lens[type()]; }

inline char *IterMap::Field::ptr()      const { return m_ptr; }
inline char *IterMap::Field::size_ptr() const { return m_ptr + 2; }
inline char *IterMap::Field::data_ptr() const { return size_ptr() + size_len(); }

inline uint32_t IterMap::Field::size() const {
  assert(isValid());

  MD::DATATYPES::TypeID t = type();
  if (t == MD::DATATYPES::STRING) {
    return MD::getint16le(size_ptr());
  } else if (t == MD::DATATYPES::BUFFER) {
    return MD::getint32le(size_ptr());
  } else {
    return MD::DATATYPES::length[t];
  }
}

inline char *IterMap::Field::next_ptr() const { return data_ptr() + size(); }
inline void  IterMap::Field::next()           { m_ptr = next_ptr(); }

inline void IterMap::getFieldIDs(std::back_insert_iterator<std::vector<FieldID> > inserter) const
{
  char *ptr     = m_data->begin() + DATA_START;
  char *end_ptr = m_data->end();

  while (ptr < end_ptr) {
    Field field(ptr);
    *inserter++ = field.id();
    ptr = field.next_ptr();
  }
}

// TODO: Make this code resilient to malformed input.  It's
// fragile and prone to buffer overflows, at the moment.
inline IterMap::Field::Field(char * ptr) : m_ptr(ptr) { }

inline uint8_t IterMap::Field::header_length() const {
  return sizeof(uint16_t) + size_len();
}

inline bool IterMap::Field::operator==(const IterMap::Field &rhs) const
{
  return (type() == rhs.type() && size() == rhs.size() &&
          !memcmp(data_ptr(), rhs.data_ptr(), size()));
}

} // namespace

#endif // __MD_LIB_ITERMAP_H__
